from idaapi import *
from idautils import *
from idc import *

def makeTail(fea, ea, notAdd = []):
	pfn = get_func(fea)
	if not pfn: return
	
	if func_contains(pfn, ea): return
	
	begin = ea
	stop = False
	toMake = []
	while not stop:
		xrefs = [xref for xref in XrefsFrom(ea)]
		if len(xrefs) == 0:
			if not isCode(getFlags(ea)):
				setFlags(ea, getFlags(ea) | FF_CODE)
			elif Byte(ea) & 0xfe != 0xc2:
				next_ea = ea + decode_insn(ea)
				if not isCode(getFlags(next_ea)):
					setFlags(next_ea, getFlags(next_ea) | FF_CODE)
			xrefs = [xref for xref in XrefsFrom(ea)]

		if len(xrefs) == 0: stop = True
		elif Byte(ea) == 0xe8:
			stop = True
			if len(xrefs) == 2:
				DelCodeXref(ea, xrefs[0].to, 0)
				del xrefs[0]
			AddCodeXref(ea, xrefs[0].to, XREF_USER | fl_JN)
			if xrefs[0].to not in notAdd: #vm1_Entry
				toMake += [xrefs[0].to]
		elif xrefs[0].type == fl_JN:
			if xrefs[0].to not in notAdd: #vm1_Entry
				toMake += [xrefs[0].to]
			stop = True
		elif len(xrefs) == 2 and (xrefs[1].type & 0x1f) >= 0x10:
			if xrefs[1].to not in notAdd: toMake += [xrefs[1].to]
			if func_contains(pfn, xrefs[0].to): stop = True
		elif xrefs[0].to in notAdd: #vm1_Entry
			stop = True

		if stop:
			end = ea + decode_insn(ea)
			break
			
		ea = xrefs[0].to
	print("Appending tail %x to func %x" %(begin, fea))
	append_func_tail(pfn, begin, end)
	for ch in toMake:
		print("Handling tail %x to append func %x" %(ch, fea))
		makeTail(fea, ch, notAdd)

def getLastInstruction(ea):
	begin = ea
	stop = False
	while not stop:
		#print "getLastInstruction(): ", ea
		xrefs = [xref for xref in XrefsFrom(ea)]
		if len(xrefs) == 0: stop = True
		elif Byte(ea) == 0xe8:
			stop = True
			if len(xrefs) == 2:
				DelCodeXref(ea, xrefs[0].to, 0)
				AddCodeXref(ea, xrefs[1].to, XREF_USER | fl_JN)
				del xrefs[0]
		elif xrefs[0].type == fl_JN:
			stop = True

		if stop:
			end = ea
			break
			
		ea = xrefs[0].to
	return end

def getBlock(ea):
	begin = ea
	stop = False
	while not stop:
		xrefs = [xref for xref in XrefsFrom(ea)]
		if len(xrefs) == 0: stop = True
		elif Byte(ea) == 0xe8:
			stop = True
			if len(xrefs) == 2:
				DelCodeXref(ea, xrefs[0].to, 0)
				AddCodeXref(ea, xrefs[1].to, XREF_USER | fl_JN)
				del xrefs[0]
		elif xrefs[0].type == fl_JN:
			stop = True

		if stop:
			end = ea + decode_insn(ea)
			break
			
		ea = xrefs[0].to
	return (begin, end)

def makeFunc(ea, notAdd = []):
	if not not get_func(ea): del_func(ea)

	fea = begin = ea
	toMake = []
	stop = False
	while not stop:
		xrefs = [xref for xref in XrefsFrom(ea)]
		if len(xrefs) == 0:
			if not isCode(getFlags(ea)):
				setFlags(ea, getFlags(ea) | FF_CODE)
			elif Byte(ea) & 0xfe != 0xc2:
				next_ea = ea + decode_insn(ea)
				if not isCode(getFlags(next_ea)):
					setFlags(next_ea, getFlags(next_ea) | FF_CODE)
			xrefs = [xref for xref in XrefsFrom(ea)]

		if len(xrefs) == 0: stop = True
		elif Byte(ea) == 0xe8:
			stop = True
			if len(xrefs) == 2:
				DelCodeXref(ea, xrefs[0].to, 0)
				del xrefs[0]
			AddCodeXref(ea, xrefs[0].to, XREF_USER | fl_JN)
			if xrefs[0].to not in notAdd: #vm1_Entry
				toMake += [xrefs[0].to]
		elif xrefs[0].type == fl_JN:
			if xrefs[0].to not in notAdd: #vm1_Entry
				toMake += [xrefs[0].to]
			stop = True
		elif len(xrefs) == 2 and (xrefs[1].type & 0x1f) >= 0x10 and xrefs[1].to not in notAdd:
			toMake += [xrefs[1].to]
		elif xrefs[0].to in notAdd: #vm1_Entry
			stop = True

		if stop:
			end = ea + decode_insn(ea)
			break
			
		ea = xrefs[0].to
	
	add_func(begin, end)
	
	for ch in toMake:
		print("Handling tail %x to append func %x" %(ch, fea))
		makeTail(fea, ch, notAdd)

	calcSpd(fea)

def calcSpd(fea):
	pfn = get_func(fea)
	enspdar = []
	nspdar = []
	spdiffar = []
	chs = [ch for ch in Chunks(fea)]
	chsaddr = []
	for ch in chs:
		cb = ch[0]
		addr = cb
		while True:
			print("tail %x" %addr)
			chsaddr += [addr]
			ce = getLastInstruction(addr)
			insn = DecodeInstruction(ce)

			if Byte(ce) == 0xe8: analyze_area(addr, ce)
			else: analyze_area(addr, ce + insn.size)

			enspd = GetSpd(ce) - GetSpd(addr)
			if Byte(ce) == 0xe8: enspd -= 4
			enspdar += [enspd]
			if len(nspdar) == 0:
				nspdar = [0]
				spdiffar = [0]
			else:
				nspdar += [-1]
				spdiffar += [-1]
			
			if ce + insn.size == ch[1]:
				break
				
			addr = ce + insn.size

	print "enspdar: ", enspdar

	if len(chsaddr) > 1:
		spdiff = 0
		spdReset = False
		chunksLesser = chsaddr[1] < fea
		for i in range(1, len(chsaddr)):
			if chunksLesser and chsaddr[i] > fea and not spdReset:
				spdReset = True
				spdiff = 0

			print("chsaddr[%x] = %x" %(i, chsaddr[i]))
			if nspdar[i] == -1:
				print("call calcSpdByBackRefs(): chsaddr[%d] = %x" %(i, chsaddr[i]))
				calcSpdByBackRefs(chsaddr, nspdar, enspdar, i, chsaddr[i])

			spdiffar[i] = nspdar[i] - GetSpd(chsaddr[i])
			spdiff += spdiffar[i]
			SetSpDiff(chsaddr[i], GetSpDiff(chsaddr[i]) + spdiffar[i])

		print "nspdar: ", nspdar
		print "spdiffar: ", spdiffar

def calcSpdByBackRefs(chsaddr, nspdar, enspdar, i, addr):
	pfn = get_func(addr)
	xrefs = [xref for xref in XrefsTo(addr, 1)]
	for xref in xrefs:
		if func_contains(pfn, xref.frm) and xref.type in [fl_JN, fl_F]:
			frmchbeg = func_tail_iterator_t(pfn, xref.frm).chunk().startEA
			frmi = chsaddr.index(frmchbeg)
			while frmi + 1 < len(chsaddr) and chsaddr[frmi + 1] > chsaddr[frmi] and chsaddr[frmi + 1] <= xref.frm: frmi += 1
			
			if nspdar[frmi] == -1:
				print("call calcSpdByBackRefs(): chsaddr[%d] = %x" %(frmi, chsaddr[frmi]))
				calcSpdByBackRefs(chsaddr, nspdar, enspdar, frmi, chsaddr[frmi])

			print "calcSpd: xref = ", hex(xref.frm), ", nspdar[", frmi, "] = ", nspdar[frmi]
			if xref.frm != getLastInstruction(chsaddr[frmi]):
				nspdar[i] = nspdar[frmi] + GetSpd(xref.frm) - GetSpd(chsaddr[frmi])
			else: nspdar[i] = nspdar[frmi] + enspdar[frmi]
			break

for f in Functions(): del_func(f)

i = 0
funcs = []
indeces = dict()
#for addr in range(0x745E3D, 0x74623D, 4):
for addr in range(0x4a68d5, 0x4a6cd5, 4):
	ea = Dword(addr) - 1
	#block = getBlock(ea)
	#print("(%x %x)" %(block[0], block[1]))
	if ea not in funcs:
		#makeFunc(ea, [0x74511E, 0x748159, 0x74512F, 0x74814E])
		makeFunc(ea, [0x4a3195, 0x4a341c, 0x4a3184, 0x4a001d])
		funcs += [ea]
		indeces[ea] = (get_name(-1, ea), [])
	indeces[ea][1].append(i)
	i += 1

for key in indeces.keys():
	print("%x %s" %(key, indeces[key]))

#makeTail(0x748403, here())